@@ -13,6 +13,9 @@ from hashlib import md5
 from itertools import product
 from functools import reduce
 from webbrowser import open as open_url
+from threading import Thread, active_count
+from time import sleep
+
 
 CONTEXT_SETTINGS = dict(max_content_width=click.get_terminal_size()[0], terminal_width=click.get_terminal_size()[0])
 
@@ -44,9 +47,10 @@ eg. cld search cat AND tags:kitten -s public_id desc -f context -f tags -n 10
 @click.option("-A", "--auto_paginate", is_flag=True, help="Return all results. Will call Admin API multiple times.")
 @click.option("-F", "--force", is_flag=True, help="Skip confirmation when running --auto-paginate")
 @click.option("-ff", "--filter_fields", multiple=True, help="Filter fields to return")
-@click.option("--save", nargs=1, help="Save output to a file")
+@click.option("--json", nargs=1, help="Save output as a JSON")
+@click.option("--csv", nargs=1, help="Save output as a CSV")
 @click.option("-d", "--doc", is_flag=True, help="Opens Search API documentation page")
-def search(query, with_field, sort_by, aggregate, max_results, next_cursor, auto_paginate, force, filter_fields, save, doc):
+def search(query, with_field, sort_by, aggregate, max_results, next_cursor, auto_paginate, force, filter_fields, json, csv, doc):
     if doc:
         open_url("https://cloudinary.com/documentation/search_api")
         exit(0)
@@ -88,19 +92,22 @@ def search(query, with_field, sort_by, aggregate, max_results, next_cursor, auto
 
     if filter_fields:
         ff = []
-        print(filter_fields)
         for f in list(filter_fields):
             if "," in f:
                 ff += f.split(",")
             else:
                 ff.append(f)
-            print(f)
         filter_fields = tuple(ff)
         all_results['resources'] = list(map(lambda x: {k: x[k] if k in x.keys() else None for k in filter_fields + with_field}, all_results['resources']))
     log(all_results)
 
-    if save:
-        write_out(all_results, save)
+    if json:
+        write_out(all_results['resources'], json)
+    
+    if csv:
+        all_results = all_results['resources']
+        #write to csv
+
 
 
 @click.command("admin",
@@ -197,8 +204,7 @@ help="""Upload a directory of assets and persist the directory structure""")
 @click.option("-f", "--folder", default="", help="Specify the folder you would like to upload resources to in Cloudinary")
 @click.option("-p", "--preset", help="Upload preset to use")
 @click.option("-v", "--verbose", is_flag=True, help="Logs information after each upload")
-@click.option("-vv", "--very_verbose", is_flag=True, help="Logs full details of each upload")
-def upload_dir(directory, optional_parameter, optional_parameter_parsed, transformation, folder, preset, verbose, very_verbose):
+def upload_dir(directory, optional_parameter, optional_parameter_parsed, transformation, folder, preset, verbose):
     items, skipped = [], []
     dir_to_upload = abspath(path_join(getcwd(), directory))
     print(f"Uploading directory '{dir_to_upload}'")
@@ -208,30 +214,45 @@ def upload_dir(directory, optional_parameter, optional_parameter_parsed, transfo
         **{k:v for k,v in optional_parameter},
         **{k:parse_option_value(v) for k,v in optional_parameter_parsed},
         "resource_type": "auto",
+        "invalidate": True,
         "unique_filename": False,
         "use_filename": True,
         "raw_transformation": transformation,
         "upload_preset": preset
     }
+
+    threads = []
+
+    def upload_multithreaded(file_path, items, skipped, v, **kwargs):
+        try:
+            _r = _uploader.upload(file_path, **kwargs)
+            print(F_OK(f"Successfully uploaded {file_path} as {_r['public_id']}"))
+            if v:
+                log(_r)
+            items.append(_r['public_id'])
+        except Exception as e:
+            print(F_FAIL(f"Failed uploading {file_path}"))
+            print(e)
+            skipped.append(file_path)
+            pass
+
     for root, _, files in walk(dir_to_upload):
         for fi in files:
             file_path = abspath(path_join(dir_to_upload, root, fi))
             mod_folder = path_join(folder, dirname(file_path[len(parent) + 1:]))
             if split(file_path)[1][0] == ".":
                 continue
-            try:
-                _r = _uploader.upload(file_path, **options, folder=mod_folder)
-                if verbose or very_verbose:
-                    print(F_OK(f"Successfully uploaded {file_path} as {_r['public_id']}"))
-                if very_verbose:
-                    log(_r)
-                items.append(_r['public_id'])
-            except Exception as e:
-                if verbose or very_verbose:
-                    print(F_FAIL(f"Failed uploading {file_path}"))
-                print(e)
-                skipped.append(file_path)
-                pass
+            options = {**options, "folder": mod_folder}
+            threads.append(Thread(target=upload_multithreaded, args=(file_path, items, skipped, verbose), kwargs=options))
+
+    for t in threads:
+        while active_count() >= 30:
+            # prevent concurrency overload
+            sleep(1)
+        t.start()
+        sleep(1/10)
+
+    for t in threads: t.join()
 
     print(F_OK(f"\n{len(items)} resources uploaded:"))
     print(F_OK('\n'.join(items)))
@@ -389,6 +410,7 @@ def sync(local_folder, cloudinary_folder, push, pull, verbose):
     skipping = 0
 
     if push:
+            
         files_to_delete_from_cloudinary = list(cld_files_ - files_)
         files_to_push = files_ - cld_files_
         files_to_check = files_ - files_to_push
@@ -423,12 +445,27 @@ def sync(local_folder, cloudinary_folder, push, pull, verbose):
 
         to_upload = list(filter(lambda x: split(x)[1][0] != ".", files_to_push))
         print("Uploading {} items to Cloudinary folder '{}'".format(len(to_upload), cloudinary_folder))
+
+        threads = []
+
+        def threaded_upload(options, path, verbose):
+            res = _uploader.upload(path, **options)
+            if verbose:
+                print(F_OK("Uploaded '{}'".format(res['public_id'])))
+
         for i in to_upload:
             modif_folder = path_join(cloudinary_folder, sep.join(i.split(sep)[:-1]))
             options = {'use_filename': True, 'unique_filename': False, 'folder': modif_folder, 'invalidate': True, 'resource_type': 'auto'}
-            res = _uploader.upload(files[i]['path'], **options)
-            if verbose:
-                print(F_OK("Uploaded '{}'".format(res['public_id'])))
+            threads.append(Thread(target=threaded_upload, args=(options, files[i]['path'], verbose)))
+        
+        for t in threads:
+            while active_count() >= 30:
+                # prevent concurrency overload
+                sleep(1)
+            t.start()
+            sleep(1/10)
+
+        [t.join() for t in threads]
 
         print("Done!")
         
@@ -485,9 +522,9 @@ def sync(local_folder, cloudinary_folder, push, pull, verbose):
 
         print("Downloading {} files from Cloudinary".format(len(files_to_pull)))
         
-        for i in files_to_pull:
-            local_path = abspath(path_join(local_folder, i + "." + cld_files[i]['format'] if cld_files[i]['resource_type'] != 'raw' else i))
-            create_required_directories(split(local_path)[0], verbose)
+        threads = []
+        
+        def threaded_pull(local_path, verbose, cld_files):
             with open(local_path, "wb") as f:
                 to_download = cld_files[i]
                 r = get(cld_url(to_download['public_id'], resource_type=to_download['resource_type'], type=to_download['type'])[0])
@@ -495,6 +532,21 @@ def sync(local_folder, cloudinary_folder, push, pull, verbose):
                 f.close()
             if verbose:
                 print(F_OK("Downloaded '{}' to '{}'".format(i, local_path)))
+
+        for i in files_to_pull:
+            local_path = abspath(path_join(local_folder, i + "." + cld_files[i]['format'] if cld_files[i]['resource_type'] != 'raw' else i))
+            create_required_directories(split(local_path)[0], verbose)
+
+            threads.append(Thread(target=threaded_pull, args=(local_path, verbose, cld_files)))
+
+        for t in threads:
+            while active_count() >= 30:
+                # prevent concurrency overload
+                sleep(1)
+            t.start()
+            sleep(1/10)
+
+        [t.join() for t in threads]
         
         print("Done!")
 
